/*
 * Copyright (c) 2018, Larry Lile <lile@users.sourceforge.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */

#include "config.h"

#include <stdio.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef STDC_HEADERS
# include <stdlib.h>
# include <stddef.h>
#else
# ifdef HAVE_STDLIB_H
#  include <stdlib.h>
# endif
#endif

#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif

#ifdef HAVE_STRING_H
# if !defined STDC_HEADERS && defined HAVE_MEMORY_H
#  include <memory.h>
# endif
# include <string.h>
#endif

#ifdef HAVE_STRINGS_H
# include <strings.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef HAVE_PWD_H
# include <pwd.h>
#endif

#ifdef HAVE_STDARG_H
# include <stdarg.h>
#endif

#ifdef HAVE_GETOPT_H
# include <getopt.h>
#endif

#ifdef HAVE_LDAP_H
# include <ldap.h>
#endif

#include "common.h"

char *image = NULL;

int debug = 0;

void version(FILE *file);
void usage(FILE *file);
int valid_shell(char *shell);

int main(int argc, char *argv[])
{
	char *ptr;

	LDAPConfig config;

	int rc = 0;
	int c;
	int opt_index;
	static char short_opt[] = "ds:hlu:v";
	static struct option long_opt[] = {
		{"help", no_argument, 0, 'h'},
		{"version", no_argument, 0, 'v'},
		{"list-shells", no_argument, 0, 'l'},
	};

	USEREntry *user = NULL;

	struct passwd pwd, *result = NULL;
	LDAP *ld;
	int err;
	char *shell = NULL;
	char *uid = NULL;

	image = argv[0];
	if (!image)
		image = "ldapchsh";
	else if ((ptr = strrchr(image, '/')))
		image = ptr + 1;

	memset(&config, 0, sizeof(LDAPConfig));

	int pwdbufsize;
	if ((pwdbufsize = sysconf(_SC_GETPW_R_SIZE_MAX)) == -1)
		abort();
	char pwdbuffer[pwdbufsize];

	while ((c =
		getopt_long(argc, argv, short_opt, long_opt,
			    &opt_index)) != -1) {
		switch (c) {
		case -1:
			break;
		case 'd':
			debug++;
			break;
		case 'l':
			valid_shell(NULL);
			exit(0);
		case 'h':
			usage(stdout);
			exit(0);
		case 's':
			shell = optarg;
			break;
		case 'u':
			uid = optarg;
			break;
		case 'v':
			version(stdout);
			exit(0);
		default:
			usage(stderr);
			exit(-1);
		}
	}

	if (argc > optind) {
		if (argc > optind + 1) {
			fprintf(stderr, "Excess argument error.\n");
			exit(1);
		}
		uid = argv[optind];
	}

	if (shell) {
		if (*shell != '/') {
			printf("%s: shell must be a full path name.\n", image);
			exit(1);
		}

		if (access(shell, F_OK) != 0) {
			printf("%s: \"%s\" does not exist.\n", image, shell);
			exit(1);
		}

		if (access(shell, X_OK) != 0) {
			printf("%s: \"%s\" is not executable.\n", image, shell);
			exit(1);
		}

		ptr = shell;
		while (*ptr != '\0') {
			if (*ptr == ',' || *ptr == ':' || *ptr == '='
			    || *ptr == '"' || *ptr == '\n') {
				printf("%s: '%c' is not allowed.\n", image,
				       *ptr);
				exit(1);
			}

			if (iscntrl(*ptr)) {
				printf
				    ("%s: Control characters are not allowed.\n",
				     image);
				exit(1);
			}
			ptr++;
		}
	} else {
		fprintf(stderr, "No shell specified.\n");
		exit(1);
	}

	if (!valid_shell(shell)) {
#ifdef ONLY_LISTED_SHELLS
		printf("%s: \"%s\" is not listed in /etc/shells.\n",
		       image, shell);
		printf("%s: Use -l option to see list.\n", image);
		exit(1);
#else
		printf
		    ("Warning: \"%s\" is not listed in /etc/shells.\n", shell);
		printf("%s: Use -l option to see list.\n", image);
#endif
	}

	get_ldap_config(&config);

	if (getpwuid_r(getuid(), &pwd, pwdbuffer, pwdbufsize, &result) != 0
	    || !result)
		abort();

	debug_msg(1, "current user pwd.pw_name = %s\n", pwd.pw_name);

	config.auth.user = pwd.pw_name;

	if (uid == NULL)
		uid = pwd.pw_name;

	debug_msg(1, "LDAPConfig config.url = %s\n", config.url);

	if (uid)
		debug_msg(1, "target user = %s\n", uid);

	if (!(ld = ldap_connect(&config)))
		exit(1);

	if ((user = get_ldap_user_entry(ld, &config, uid)) == NULL) {
		printf("Unable to find user \"%s\" in LDAP\n", uid);
		rc = 1;
		goto done;
	}

	debug_msg(1, "user dn: %s\n", user->dn);

	if (strcmp(user->loginShell, shell) == 0) {
		printf("Shell not changed.\n");
		goto done;
	}

	char *loginshellvals[2] = { shell, NULL };

	LDAPMod loginshell = {
		.mod_op = LDAP_MOD_REPLACE,
		.mod_type = "loginShell",
		.mod_values = loginshellvals,
	};
	LDAPMod *mods[2] = { &loginshell, NULL };

	debug_msg(2, "ldap_modify_ext_s(%s, %s => %s, NULL, NULL)\n", user->dn,
		    mods[0]->mod_type, mods[0]->mod_values[0]);
	if ((err =
	     ldap_modify_ext_s(ld, user->dn, mods, NULL,
			       NULL)) != LDAP_SUCCESS) {
		fprintf(stderr, "ldap_modify_ext_s(): %d %s\n", err,
			ldap_err2string(err));
	}

 done:
	if (user)
		free_ldap_user_entry(user);

	debug_msg(2, "ldap_unbind_ext_s()\n");
	ldap_unbind_ext_s(ld, NULL, NULL);

	exit(rc);
}

void usage(FILE *file)
{
	fprintf(file,
		"Usage: %s [ -s shell ] [ --list-shells ] [ --help ] [ --version ] "
		" [ -u username | username ]\n", image);
}

int valid_shell(char *shell)
{
	FILE *file;
	char *buffer;
	char *ptr;
	size_t bufsize = 0;
	size_t count;
	int rc = 0;
#ifdef ETC_SHELLS
	char shells[] = ETC_SHELLS;
#else
	char shells[] = "/etc/shells";
#endif

	if (!(file = fopen(shells, "r"))) {
		printf("No known shells\n");
		return (rc);
	}

	while ((count = getline(&buffer, &bufsize, file)) != -1) {
		if (*buffer == '#')
			continue;

		if ((ptr = strchr(buffer, '\n')))
			*ptr = '\0';

		if (shell != NULL) {
			if (strcmp(shell, buffer) == 0) {
				rc = 1;
				break;
			}
		} else
			printf("%s\n", buffer);
	}
	free(buffer);
	fclose(file);

	return (rc);
}

void version(FILE *file)
{
	fprintf(file, PACKAGE " " PACKAGE_VERSION "\n");
}

